(function(context, undefined) {
   /**
    * Check an object is a boolean or not.
    */
   function isBoolean(obj) {
      return Object.prototype.toString.apply(obj) === "[object Boolean]";
   }

   /**
    * Check an object is a number or not.
    */
   function isNumber(obj) {
      return Object.prototype.toString.apply(obj) === "[object Number]";
   }

   /**
    * Check an object is empty or not. An empty object is either null, undefined, "", [], or {}.
    * A string contains blank spaces (blank, tab, etc.) also is treated as an empty string.
    */
   function isEmpty(obj) {
      return (obj === undefined || obj === null || (isString(obj) && trim(obj) === "") || (isArray(obj) && obj.length === 0) || (isMap(obj) && (function(ob) { for (var i in ob) { return false; } return true; })(obj)));
   }

   /**
    * Get actual object from its name.
    */
   function fromString(name) {
      var names = name.split("."), root = context, i = 0, len = names.length - 1;

      while (i <= len) {
         root = root[names[i]];
         if ( !root) {
            return root;
         }
         i++;
      }
      return root;
   }

   /**
    * Trim a string. If str is not a string, the whole object will be returned.
    * @see http://tinyurl.com/yl934kz
    */
   function trim(str) {
      var chars = " \n\r\t\v\f\u00a0\u2000\u2001\u2002\u2003\u2004\u2005\u2006\u2007\u2008\u2009\u200a\u200b\u2028\u2029\u3000",
          ws    = {}, len = chars.length;

      while (len--) {
         ws[chars.charAt(len)] = true;
      }

      trim = function(str) {
         if ( !isString(str)) {
            return str;
         }
         if (String.prototype.trim) {
            return String.prototype.trim.apply(this, str);
         }
         var s = -1, e = str.length;
         while (ws[str.charAt(--e)]) {};
         while (s++ !== e && ws[str.charAt(s)]) {};
         return str.substring(s, e + 1);
      };
      return trim(str);
   }

   /*
    * Make overloading method in case api is declared as an array.
    */
   function makeOverloadingArray(methodName, api) {
      // It is an array, check the length
      if (api.length === 0) {
         return api;
      }

      // Ok, it is an array, check if it is not a method but a normal array property
      for (var idx in api) {
         if ( !isFunction(api[idx])) {
            return api;
         }
      }

      // Now it is an overloading method. If there is only one overloading, ignore wrapping
      if (api.length === 1) {
         return api[0];
      }

      // Make wrapper
      return function() {
         var len = arguments.length, defaultFn = null, idx;

         for (idx in api) {
            if (api[idx].length === len) {
               return api[idx].apply(this, arguments);
            } else if (api[idx].length === 0) {
               defaultFn = api[idx];
            }
         }

         // Still not match? How about passing to the default method which does not declare any argument?
         if (defaultFn) {
            return defaultFn.apply(this, arguments);
         } else { // No method supports arguments.length? Throw an exception
            throw "overload: " + methodName + "() does not accept " + len + " arguments";
         }
      };
   }

   /*
    * Make overloading method in case api is declared as an object.
    * @param {String} methodName method name.
    * @param {Object} api api declared as an object.
    * @return {Function} function supports all declared overloading.
    * @throw {Exception} if there is an error in an overloading method.
    * @private
    */
   function makeOverloadingObject(methodName, api) {
      /*
       * Make evaluator function for an argument of a overriding function.
       */
      function makeEvaluator(expression, typeDefsString) {
         var s = "(function() { return function(it) { return (" + expression + ") === true; }})();";
         try {
            return eval(s);
         } catch (error) {
            throw "overload: Invalid validating expression: " + expression + " on overloading method " + methodName + "(" + typeDefsString + ")";
         }
      }

      /*
       * Transform types definition for a overriding function.
       * typeDefs ~ ["string", "string: it != null"]
       * >>: [{ name: "string", type: String },
       *       { name: "String", type: String, expression: "it != null", evaluator: function(it) { return (it != null) === true; }]
       * @param typeDefs array of type definitions.
       * @param typeDefsString the whole string represent types definition of the method.
       */
      function transformTypes(typeDefs, typeDefsString) {
         var result = [];

         each(typeDefs, function(typeDef, index) {
            typeDef = trim(typeDef);
            var name, type, expression, evaluator, colon = typeDef.indexOf(":");

            if (colon === -1) {
               name = typeDef;
            } else {
               name = typeDef.substring(0, colon);
               expression = typeDef.substring(colon + 1, typeDef.length);
               evaluator = makeEvaluator(expression, typeDefsString);
            }
            try {
               type = eval(name);
            } catch (error) {
               throw "overload: Type " + name + " is not defined on overloading method " + methodName + "(" + typeDefsString + ")";
            }
            result.push({
               name: name,
               type: type,
               expression: expression,
               evaluator: evaluator
            });
         });
         return result;
      }

      /*
       * Determine type of an argument.
       */
      function getType(obj) {
         if (isString(obj))   { return String; }
         if (isArray(obj))    { return Array; }
         if (isNumber(obj))   { return Number; }
         if (isFunction(obj)) { return Function; }
         if (isBoolean(obj))  { return Boolean; }

         // Try to guess obj is a class created by jsface
         try {
            if (isString(obj.$meta.name)) {
               return eval(obj.$meta.name); // TODO Optimize this
            }
         } catch (error) {}
         return Object; // all other objects will be marked as Object type
      }

      /*
       * Check that arguments passing to call function matches an overriding function or not.
       */
      function isMatched(args, len, typeInfo) {
         for (var i = 0, type, name, arg; i < len; i++) {
            type = typeInfo[i].type;
            name = typeInfo[i].name;
            arg  = args[i];

            if (getType(arg) !== type && !(arg === null && (type === String || type === Array || type === Object))) {
               return false;
            }
         }
         return true;
      }

      /*
       * Build method signature for debugging.
       */
      function buildMethodSignature(typeItem, args, len) {
         for (var s = methodName + "(", i = 0; i < len - 1; i++) {
            s += typeItem.types[i].name + ": " + args[i] + ", ";
         }
         s += typeItem.types[len - 1].name + ": " + args[len - 1] + ")";
         return s;
      }

      /*
       * Select an overriding function to execute function with passing arguments.
       */
      function selectHandler(ctx, mapItem, len, args) {
         var matches = [], typeItem, i, s;

         // Iterator over mapItem, if arguments match types definition, push matched meta in matches
         each(mapItem, function(item, index) {
            if (isMatched(args, len, item.types)) {
               matches.push(item);
            }
         });

         // There are more than one methods matched, check to remove methods which arguments do not pass evaluators
         if (matches.length > 1) {
            each(matches, function(typeItem, index) {
               for (i = 0; i < len; i++) {
                  var evaluator = typeItem.types[i].evaluator;
                  if (evaluator) {
                     if (evaluator.call(ctx, args[i]) === false) {
                        matches.splice(index, 1);
                     }
                  }
               }
            });
         }

         // Only one method matched, great, it is the method for arguments
         if (matches.length === 1) {
            typeItem = matches[0];

            for (i = 0; i < len; i++) {
               var evaluator = typeItem.types[i].evaluator;
               if (evaluator) {
                  if (evaluator.call(ctx, args[i]) !== true) {
                     throw buildMethodSignature(typeItem, args, len) +
                        ". Validating error at parameter " + (i + 1) + ", expression: " +
                        typeItem.types[i].expression;
                  }
               }
            }
            return typeItem.fn.apply(ctx, args);
         }

         // Prepare error message
         s = methodName + "(";
         for (i = 0; i < len - 1; i++) {
            s += args[i] + ", ";
         }
         s += args[len - 1] + "). Check argument types and values.";

         // Something wrong here
         if (matches.length === 0) {
            throw "No overloading method matches the call " + s;
         } else {
            throw "Vague arguments on calling " + s;
         }
      }

      // Where to store all information about the method
      var overloadingsMeta = {};

      // Push default overloadingsMeta
      if (isFunction(api["0"])) {
         overloadingsMeta["0"] = [{
            fn: api["0"]
         }];
      }

      // Pre-check, and process types during the checking time
      for (var key in api) {
         // Check for sure that the whole contents are key:function
         if ( !isFunction(api[key])) {
            return api;
         }

         // Skip checking default handler
         if (key === "0") {
            continue;
         }

         var typeDefs = key.split(","), f = api[key];

         // If there is only one typeDef, and it's not a type, skip processing: return AS IS
         if (typeDefs.length === 1) {
            if ( !fromString(typeDefs[0].split(":")[0])){
               return api;
            }
         }

         // Check, types length != function parameters length
         if (typeDefs.length !== f.length) {
            throw "Invalid method declaration for " + methodName + "() at overloading \"" + key +
               "\". Actual overloading parameters do not match with their types declaration.";
         }

         // Prepare overloadingsMeta[] for this function
         if (overloadingsMeta[typeDefs.length] === undefined) {
            overloadingsMeta[typeDefs.length] = [];
         }

         // Transform typeDefs from array of string to array of object
         typeDefs = transformTypes(typeDefs, key);

         // Put method info (type declaration and function) into overloadingsMeta
         overloadingsMeta[typeDefs.length].push({ types: typeDefs, fn: f });
      }

      // Make wrapper
      return function(overloadings) {
         return function() {
            var len = arguments.length;

            if (overloadings[len] !== undefined) {         // Match, delegate to selectHandler
               return selectHandler(this, overloadings[len], len, arguments);
            } else if (overloadings[0] !== undefined) {    // Default handler
               return overloadings[0][0].fn.apply(this, arguments);
            } else {                                       // No one matched, throw an exception
               throw methodName + "() does not accept " + len + " arguments";
            }
         };
      }(overloadingsMeta);
   }

   /**
    * Make overloading methods.
    */
   function overload(methodName, api) {
      if ( !isEmpty(api)) {
         if (isArray(api)) {
            return makeOverloadingArray(methodName, api);
         } else if (isMap(api)) {
            return makeOverloadingObject(methodName, api);
         }
      }
      return api;
   }

   if (typeof module !== "undefined" && module.exports) {
      var jsface = require("./jsface");
      jsface.Class.overload = overload;
   }

   var extend     = jsface.extend,
       each       = jsface.each,
       isMap      = jsface.isMap,
       isArray    = jsface.isArray,
       isFunction = jsface.isFunction,
       isString   = jsface.isString,
       isClass    = jsface.isClass;
})(typeof global !== "undefined" ? global : this);